package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** an element which can produce one of several alternatives */
public class Union extends Element implements Iterable<Sequence> {

    final String shortForm;
    final boolean synthetic;
    private final List<Sequence> alternatives = new ArrayList<Sequence>();

    public Iterator<Sequence> iterator() { return alternatives.iterator(); }

    void reachable(HashSet<Sequence.Position> h) { for(Sequence s : alternatives) s.reachable(h); }

    /** adds an alternative */
    public void add(Sequence s) { alternatives.add(s); }

    /**
     *  Since every cycle in a non-degenerate grammar contains at
     *  least one Union, every instance of this class must be able to
     *  display itself in both "long form" (list of the long forms of
     *  its alternatives) and "short form" (some abbreviation).
     *
     *  @param shortForm the "short form" display; usually 
     *  @param synthetic if true, this Union's "long form" is "obvious" and should not be displayed when printing the grammar
     */
    public Union(String shortForm) { this(shortForm, false); }
    public Union(String shortForm, boolean synthetic) {
        this.shortForm = shortForm;
        this.synthetic = synthetic;
    }

    private Forest.Ref epsilonForm = null;
    Forest epsilonForm() {
        if (epsilonForm != null) return epsilonForm;
        epsilonForm = new Forest.Ref();
        for(Sequence s : this)
            if (s.possiblyEpsilon(null))
                epsilonForm.merge(s.epsilonForm());
        return epsilonForm;
    }

    // Display //////////////////////////////////////////////////////////////////////////////

    public String toString() { return shortForm; }
    private static String pad(int i,String s) { return s.length() >= i ? s : pad(i-1,s)+" "; }
    void toString(StringBuffer sb) {
        if (synthetic) return;
        boolean first = true;
        if (alternatives.size()==0) {
            sb.append(pad(15, shortForm) + "::= ");
        } else for(Sequence s : this) {
            sb.append(pad(15, first ? shortForm : "") + (first ? "::= " : "  | "));
            first = false;
            sb.append(s.toString());
            sb.append('\n');
        }
        sb.append('\n');
    }

    // SubUnion //////////////////////////////////////////////////////////////////////////////

    /** FIXME this is kind of a hack */
    public class Subset extends Union {
        private final Set<Sequence> exclude;
        public Subset(String shortForm, Set<Sequence> exclude) { super(shortForm, true); this.exclude = exclude; }
        public Iterator<Sequence> iterator() {
            final Iterator<Sequence> it = Union.this.iterator();
            return new Iterator<Sequence>() {
                private Sequence next = it.hasNext() ? it.next() : null;
                public boolean hasNext() { return next != null; }
                public Sequence next() {
                    Sequence ret = next;
                    do {
                        next = it.hasNext() ? it.next() : null;
                    } while (next != null && (next instanceof Sequence) && exclude.contains((Sequence)next));
                    return ret;
                }
                public void remove() { throw new Error(); }
            };
        }
    }

}
